unit hStrain_main;

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, ComCtrls, Menus, Math, TeEngine, Series, TeePolar, TeeTools,
  ExtCtrls, TeeProcs, Chart, ExtDlgs, StdCtrls, Grids, TeeSurfa, TeePoin3, About;

const
  NumBootstrap    =  200;
  ConfidenceLevel = 0.95;

type
  Thyperbolic_vector = array[0..2] of single;
  Thyperbolic_matrix = array[0..2,0..2] of single;
  Tvector2           = array[1..2] of single;

  Tdata = record
    a, b, R, ph : single;
    rh, zt, ps  : single;
    x           : Thyperbolic_vector;
  end;

  Tbootstrapsolution = record
    rh, ps : single;
    x      : Thyperbolic_vector;
    ds     : single; // hyperbolic distance from Xs
  end;

  TFormMain = class(TForm)
    MainMenu1: TMainMenu;
    MenuFile: TMenuItem;
    MenuOpen: TMenuItem;
    N1: TMenuItem;
    MenuExit: TMenuItem;
    MenuEdit: TMenuItem;
    MenuHelp: TMenuItem;
    MenuVersion: TMenuItem;
    StatusBar1: TStatusBar;
    PageControl1: TPageControl;
    TabSheetPlot: TTabSheet;
    TabSheetValues: TTabSheet;
    GridBandTool1: TGridBandTool;
    Chart1: TChart;
    SeriesB: TPolarSeries;
    SeriesF: TPolarSeries;
    SeriesI: TPolarSeries;
    SeriesO: TPolarSeries;
    GridBandTool2: TGridBandTool;
    SavePictureDialog1: TSavePictureDialog;
    Memo1: TMemo;
    SeriesFh: TPolarSeries;
    SeriesIh: TPolarSeries;
    StringGrid1: TStringGrid;
    OpenDialog1: TOpenDialog;
    SaveDialog1: TSaveDialog;
    MenuSaveChart: TMenuItem;
    procedure MenuExitClick(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure Chart1MouseUp(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure Chart1Resize(Sender: TObject);
    procedure MenuOpenClick(Sender: TObject);
    procedure MenuSaveChartClick(Sender: TObject);
    procedure Chart1MouseMove(Sender: TObject; Shift: TShiftState; X,
      Y: Integer);
    procedure StringGrid1MouseDown(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure MenuVersionClick(Sender: TObject);
  private
    Ndat : integer;
    DatF : array of Tdata;
    DatI : array of Tdata;
    Dats : Tdata;
    Bsols: TList; // bootstrap solutions
    Bs   : array[1..NumBootstrap] of Tbootstrapsolution;
    Q_I2F : Thyperbolic_matrix; // transformation matrix for straining
    Q_F2I : Thyperbolic_matrix; // transformation matrix for destraining
    IsoRcircleCount : integer;
    IsoRcircle : array of TPolarSeries;
    procedure SeriesAllClear();
    procedure DrawIsoRcircles(const Rmax: single);
    procedure EraseIsoRcircles();
    function  ReadData(): boolean;
    function  FRmax(): single;
    procedure Rfphi2hypcoords(const k: AnsiChar);
    procedure PlotBsols();
    procedure PlotDatF();
    procedure PlotDatI();
    procedure CalcDatI();
    procedure MeanStrain();
    procedure ComposeTransformMatrices();
    procedure Bootstrap();
  public
  end;

var
  FormMain: TFormMain;
  function R2rho(const R: single): single;
  function rho2zeta(const rho: single): single;
  function R2zeta(const R: single): single;
  function rhopsi2x(const rho, psi: single): Thyperbolic_vector;
  function hyperbolic_dist(const a, b: Thyperbolic_vector): single;
  function hyperbolic_dotprod(const a, b: Thyperbolic_vector): single;
  function hyperbolic_eye(): Thyperbolic_matrix;
  function hyperbolic_matdet(const a: Thyperbolic_matrix): single;
  function hyperbolic_matinv(const a: Thyperbolic_matrix): Thyperbolic_matrix;
  function hyperbolic_matmul(const a, b: Thyperbolic_matrix): Thyperbolic_matrix;
  function hyperbolic_matvecmul(const a: Thyperbolic_matrix; const v: Thyperbolic_vector): Thyperbolic_vector;
  function hyperbolic_norm(const v: Thyperbolic_vector): single;
  function hyperbolic_zeromat(): Thyperbolic_matrix;
  procedure x2rhopsi(const x: Thyperbolic_vector; var rho, psi: single);
  function CompareDistances(b1, b2: pointer): integer;
  function vecprod(const a, b: Tvector2): single;
  function normalize_vector2(const a: Tvector2; var n: single): Tvector2;

implementation

{$R *.dfm}
////////////////////////////////////////////////////////////////////////////////

function CompareDistances(b1, b2: pointer): integer;
begin
  if      Tbootstrapsolution(b1^).ds < Tbootstrapsolution(b2^).ds then Result := -1
  else if Tbootstrapsolution(b1^).ds > Tbootstrapsolution(b2^).ds then Result :=  1
  else Result := 0;
end; {CompareDistances}

//------------------------------------------------------------------------------
function hyperbolic_dist(const a, b: Thyperbolic_vector): single;
begin
  Result := ArcCosh(-hyperbolic_dotprod(a,b));
end; {function hyperbolic_dist}

//------------------------------------------------------------------------------
function hyperbolic_dotprod(const a, b: Thyperbolic_vector): single;
begin
  Result := -a[0]*b[0] + a[1]*b[1] + a[2]*b[2];
end; {function hyperbolic_dotprod}

//------------------------------------------------------------------------------
function hyperbolic_eye(): Thyperbolic_matrix;
begin
  Result[0,0] := 1; Result[0,1] := 0; Result[0,2] := 0;
  Result[1,0] := 0; Result[1,1] := 1; Result[1,2] := 0;
  Result[2,0] := 0; Result[2,1] := 0; Result[2,2] := 1;
end; {function hyperbolic_eye}

//------------------------------------------------------------------------------
function hyperbolic_matdet(const a: Thyperbolic_matrix): single;
begin
  Result := a[0,0]*a[1,1]*a[2,2] + a[0,2]*a[2,1]*a[1,0] + a[2,0]*a[0,1]*a[1,2]
          - a[2,0]*a[1,1]*a[0,2] - a[0,0]*a[2,1]*a[1,2] - a[2,2]*a[1,0]*a[0,1];
end; {function hyperbolic_matdet}

//------------------------------------------------------------------------------
function hyperbolic_matinv(const a: Thyperbolic_matrix): Thyperbolic_matrix;
var
  i, j : byte;
  det  : single;
begin
  Result[0,0] := a[1,1]*a[2,2]-a[1,2]*a[2,1];
  Result[0,1] := a[2,1]*a[0,2]-a[2,2]*a[0,1];
  Result[0,2] := a[0,1]*a[1,2]-a[0,2]*a[1,1];

  Result[1,0] := a[1,2]*a[2,0]-a[1,0]*a[2,2];
  Result[1,1] := a[2,2]*a[0,0]-a[2,0]*a[0,2];
  Result[1,2] := a[0,2]*a[1,0]-a[0,0]*a[1,2];

  Result[2,0] := a[1,0]*a[2,1]-a[1,1]*a[2,0];
  Result[2,1] := a[2,0]*a[0,1]-a[2,1]*a[0,0];
  Result[2,2] := a[0,0]*a[1,1]-a[0,1]*a[1,0];

  det := hyperbolic_matdet(a);
  for i := 0 to 2 do for j := 0 to 2 do Result[i,j] := Result[i,j] / det;
end; {function hyperbolic_matinv}

//------------------------------------------------------------------------------
function hyperbolic_matmul(const a, b: Thyperbolic_matrix): Thyperbolic_matrix;
var i, j, k : byte;
begin
  Result := hyperbolic_zeromat();
  for i := 0 to 2 do for j := 0 to 2 do for k := 0 to 2 do
    Result[i,j] := Result[i,j] + a[i,k]*b[k,j];
end; {function hyperbolic_matmul}

//------------------------------------------------------------------------------
function hyperbolic_matvecmul(const a: Thyperbolic_matrix; const v: Thyperbolic_vector): Thyperbolic_vector;
var i, j : byte;
begin
  for i := 0 to 2 do Result[i] := 0;
  for i := 0 to 2 do for j := 0 to 2 do Result[i] := Result[i] + a[i,j]*v[j];
end; {function hyperbolic_matvecmul}

//------------------------------------------------------------------------------
function hyperbolic_norm(const v: Thyperbolic_vector): single;
begin
  Result := sqrt(-hyperbolic_dotprod(v,v));
end; {function hyperbolic_norm}

//------------------------------------------------------------------------------
function hyperbolic_zeromat(): Thyperbolic_matrix;
var i, j : byte;
begin
  for i := 0 to 2 do for j := 0 to 2 do Result[i,j] := 0;
end; {function hyperbolic_zeromat}

//------------------------------------------------------------------------------
function normalize_vector2(const a: Tvector2; var n: single): Tvector2;
begin
  n := sqrt( sqr(a[1])+sqr(a[2]) );
  if n=0 then begin
    Result[1] := 0.0;
    Result[2] := 0.0;
  end else begin
    Result[1] := a[1]/n;
    Result[2] := a[2]/n;
  end; {if}
end; {function normalize_vector2}

//------------------------------------------------------------------------------
function R2rho(const R: single): single;
begin
  Result := ln(R);
end; {function R2rho}

//------------------------------------------------------------------------------
function rho2zeta(const rho: single): single;
begin
  Result := 2 * sinh(rho/2);
end; {function rho2zeta}

//------------------------------------------------------------------------------
function R2zeta(const R: single): single;
begin
  Result := sqrt(R) - 1.0/sqrt(R);
end; {function R2zeta}

//------------------------------------------------------------------------------
function rhopsi2x(const rho, psi: single): Thyperbolic_vector;
begin
  Result[0] := cosh(rho);
  Result[1] := sinh(rho) * cos(psi);
  Result[2] := sinh(rho) * sin(psi);
end; {function rhopsi2x}

//------------------------------------------------------------------------------
function vecprod(const a, b: Tvector2): single;
begin
  Result := a[1]*b[2] - a[2]*b[1];
end; {function vecprod}

//------------------------------------------------------------------------------
procedure x2rhopsi(const x: Thyperbolic_vector; var rho, psi: single);
begin
  rho := ArcSinh(sqrt(sqr(x[1])+sqr(x[2])));
  psi := ArcTan2(x[2],x[1]);
end; {procedure x2rhopsi}

////////////////////////////////////////////////////////////////////////////////

//------------------------------------------------------------------------------
procedure TFormMain.Bootstrap();
var
  bset_rh, bset_ps : array of single;
  brh, bps         : single;
  x, v : Thyperbolic_vector;
  s    : array[0..2] of double;
  ib   : integer;
  id   : integer;
  ir   : integer;
  i, j : byte;
begin
  SetLength(bset_rh, Ndat+1);
  SetLength(bset_ps, Ndat+1);
  RandSeed := 1;

  for ib := 1 to NumBootstrap do begin
    for id := 1 to Ndat do begin // bootstrap set
      ir := ceil(random*Ndat);
      bset_rh[id] := DatF[ir].rh;
      bset_ps[id] := DatF[ir].ps;
    end; {for}
    for i := 0 to 2 do s[i] := 0;
    for id := 1 to Ndat do begin
      x := rhopsi2x(bset_rh[id],bset_ps[id]);
      for i := 0 to 2 do s[i] := s[i] + x[i];
    end; {for}
    for i := 0 to 2 do v[i]  := s[i];
    for i := 0 to 2 do Bs[ib].x[i] := s[i] / hyperbolic_norm(v);
    Bs[ib].ds := hyperbolic_dist(Bs[ib].x, Dats.x);
  end; {for ib}

  Finalize(bset_rh);
  Finalize(bset_ps);
end; {procedure TFormMain.Bootstrap}

//------------------------------------------------------------------------------
procedure TFormMain.CalcDatI();
var
  i : integer;
begin
  for i := 1 to Ndat do begin
    DatI[i].x := hyperbolic_matvecmul(Q_F2I,DatF[i].x);
    x2rhopsi(DatI[i].x, DatI[i].rh, DatI[i].ps );
    DatI[i].zt := rho2zeta(DatI[i].rh);
    DatI[i].R  := exp(DatI[i].rh);
    DatI[i].ph := RadToDeg(DatI[i].ps/2);
  end; {for}
end; {procedure TFormMain.CalcDatI}

//------------------------------------------------------------------------------
procedure TFormMain.Chart1MouseMove(Sender: TObject; Shift: TShiftState; X,
  Y: Integer);
begin
  {}
end; {procedure TFormMain.Chart1MouseMove}

//------------------------------------------------------------------------------
procedure TFormMain.Chart1MouseUp(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
var
  fname : string;
  ex    : string;
begin
  if Button = mbRight then
    if SavePictureDialog1.Execute then begin
      fname := SavePictureDialog1.FileName;
      if FileExists(fname) then
        if MessageDlg('File already exist. Overwrite?',
           mtWarning, [mbYes, mbNo], 0) = mrNo then Exit;     //===>
      ex := UpperCase(ExtractFileExt(fname));
      if      ex = '.EMF' then Chart1.SaveToMetafileEnh(fname)
      else if ex = '.BMP' then Chart1.SaveToBitmapFile( fname)
      else begin
        ShowMessage('Invalid file extension.');
        Exit;                                                 //===>
      end; {if}
  end; {if Button and if SavePictureDialog1}

  if Button = mbLeft then begin
    SeriesFh.Visible := False;
    SeriesIh.Visible := False;
  end; {if}
end; {procedure TFormMain.Chart1MouseUp}

//------------------------------------------------------------------------------
procedure TFormMain.Chart1Resize(Sender: TObject);
var i : integer;
begin
  with StringGrid1 do begin
    Left := Chart1.Width - Width;
    Top  := max(112, Chart1.Height div 5);
    Height := Chart1.Height - Top;
    RowCount := Height div DefaultRowHeight;
    for i := 1 to RowCount do Cells[0,i] := IntToStr(i);
    Cells[1,0] := 'major';
    Cells[2,0] := 'minor';
    Cells[3,0] := 'angle';
    Cells[4,0] := 'R';
  end; {with}
end; {procedure TFormMain.Chart1Resize}

//------------------------------------------------------------------------------
procedure TFormMain.ComposeTransformMatrices();
var
  a, b, c : Thyperbolic_matrix;
begin
  a := hyperbolic_eye();
  b := hyperbolic_eye();
  c := hyperbolic_eye();
  a[1,1] := cos(Dats.ps);  a[1,2] := -sin(Dats.ps);
  a[2,1] := sin(Dats.ps);  a[2,2] :=  cos(Dats.ps);
  b[0,0] := cosh(Dats.rh); b[0,1] := sinh(Dats.rh);
  b[1,0] := sinh(Dats.rh); b[1,1] := cosh(Dats.rh);
  c[1,1] := cos(Dats.ps);  c[1,2] :=  sin(Dats.ps);
  c[2,1] :=-sin(Dats.ps);  c[2,2] :=  cos(Dats.ps);
  Q_I2F  := hyperbolic_matmul(hyperbolic_matmul(a,b),c);
  Q_F2I  := hyperbolic_matinv(Q_I2F);
end; {procedure TFormMain.ComposeTransformMatrices}

//------------------------------------------------------------------------------
procedure TFormMain.DrawIsoRcircles(const Rmax: Single);
var
  maxR : integer;
  iR   : integer;
  zeta : single;
  ipsi : integer;
begin
  maxR := ceil(Rmax);
  IsoRcircleCount := maxR-1;
  Finalize(  IsoRcircle );
  SetLength( IsoRcircle, IsoRcircleCount );

  for iR := 2 to maxR do begin
    zeta := sqrt(iR) - 1.0/sqrt(iR);
    IsoRcircle[iR-2] := TPolarSeries.Create(Self);
    with IsoRcircle[iR-2] do begin
      ParentChart := Chart1;
      Clear();
      Circled := True;
      Pointer.Style := psNothing;
      ShowInLegend := False;
      IsoRcircle[iR-2].Pen.Color := clGray;
      for ipsi := 0 to 180 do AddPolar(2*ipsi, zeta);
      if iR = maxR then begin AddPolar(0,0); AddPolar(0,zeta); end;
    end; {with}
  end; {for iR}
end; {procedure TFormMain.DrawIsoRcircles}

//------------------------------------------------------------------------------
procedure TFormMain.EraseIsoRcircles();
var
  i : integer;
begin
  for i := 0 to IsoRcircleCount-1 do begin
    IsoRcircle[i].Clear();
    IsoRcircle[i].Destroy();
  end; {for}
  Finalize( IsoRcircle );
end; {procedure TFormMain.EraseIsoRcircles}

//------------------------------------------------------------------------------
procedure TFormMain.FormCreate(Sender: TObject);
const R_outercircle = 4;
begin
  SetLength( IsoRcircle, 2 );
  DrawIsoRcircles(R_outercircle);
  SeriesAllClear();
  Memo1.Lines.Clear();
  SetLength(DatF,2);
  SetLength(DatI,2);
  SeriesFh.Transparency := 0;
  SeriesIh.Transparency := 0;
  SeriesFh.Visible := False;
  SeriesIh.Visible := False;
  Bsols := TList.Create();
  Bsols.Clear();
end; {procedure TFormMain.FormCreate}

//------------------------------------------------------------------------------
procedure TFormMain.FormDestroy(Sender: TObject);
begin
  EraseIsoRcircles();
  Finalize( IsoRcircle );
  Finalize( DatF );
  Finalize( DatI );
  Bsols.Free();
end; {procedure TFormMain.FormDestroy}

//------------------------------------------------------------------------------
function TFormMain.FRmax(): single;
var
  i : integer;
begin
  Result := 0;
  for i := 1 to Ndat do
    Result := max( max( Result, DatF[i].R), DatI[i].R );
end; {function TFormMain.Fzetamax}

//------------------------------------------------------------------------------
procedure TFormMain.MeanStrain();
var
  s  : array[0..2] of double;
  v  : Thyperbolic_vector;
  i : integer;
  j : byte;
begin
  for j := 0 to 2 do s[j] := 0;
  for i := 1 to Ndat do for j := 0 to 2 do s[j] := s[j] + DatF[i].x[j];
  for j := 0 to 2 do v[j] := s[j];
  for j := 0 to 2 do Dats.x[j] := s[j] / hyperbolic_norm(v);

  x2rhopsi(Dats.x, Dats.rh, Dats.ps);
  Dats.zt := 2.0 * sinh(Dats.rh/2.0);
  Dats.R  := exp(Dats.rh);
  Dats.ph := RadToDeg(Dats.ps / 2.0);
  SeriesO.Clear();
  SeriesO.AddPolar(RadToDeg(Dats.ps), Dats.zt);
  SeriesO.Pen.Visible := True;

  Memo1.Lines.Add('');
  Memo1.Lines.Add('*** Optimal Strain ***');
  Memo1.Lines.Add('  R_s    = ' + FloatToStrF(Dats.R, ffFixed, 6,3));
  Memo1.Lines.Add('  phi_s  = ' + FloatToStrF(Dats.ph,ffFixed, 6,3));
  Memo1.Lines.Add('  rho_s  = ' + FloatToStrF(Dats.rh,ffFixed, 6,3));
  Memo1.Lines.Add('  zeta_s = ' + FloatToStrF(Dats.zt,ffFixed, 6,3));
  Memo1.Lines.Add('  psi_s  = ' + FloatToStrF(RadToDeg(Dats.ps),ffFixed, 6,3));

end; {procedure TFormMain.MeanStrain}

//------------------------------------------------------------------------------
procedure TFormMain.MenuExitClick(Sender: TObject);
begin
  Application.Terminate();
end; {procedure TFormMain.MenuExitClick}

//------------------------------------------------------------------------------
procedure TFormMain.MenuOpenClick(Sender: TObject);
var i : integer;
begin
  if not ReadData() then Exit;           //===>
  Rfphi2hypcoords('F');
  EraseIsoRcircles();
  // DrawIsoRcircles( FRmax() );
  PlotDatF();
  for i := 1 to Ndat do DatF[i].x := rhopsi2x(DatF[i].rh, DatF[i].ps);

  MeanStrain();
  ComposeTransformMatrices();
  CalcDatI();
  PlotDatI();

  DrawIsoRcircles( FRmax() );

  if Ndat < 5 then begin
    ShowMessage('Number of data < 5');
    Exit;                                //===>
  end; {if}

  Bootstrap();
  PlotBsols();
end; {procedure TFormMain.MenuOpenClick}

//------------------------------------------------------------------------------
procedure TFormMain.MenuSaveChartClick(Sender: TObject);
var
  fname : string;
  ex    : string;
begin
    if SavePictureDialog1.Execute then begin
      fname := SavePictureDialog1.FileName;
      if FileExists(fname) then
        if MessageDlg('File already exist. Overwrite?',
           mtWarning, [mbYes, mbNo], 0) = mrNo then Exit;     //===>
      ex := UpperCase(ExtractFileExt(fname));
      if      ex = '.EMF' then Chart1.SaveToMetafileEnh(fname)
      else if ex = '.BMP' then Chart1.SaveToBitmapFile( fname)
      else begin
        ShowMessage('Invalid file extension.');
        Exit;                                                 //===>
      end; {if}
  end; {if Button and if SavePictureDialog1}
end; {procedure TFormMain.MenuSaveChartClick}

//------------------------------------------------------------------------------
procedure TFormMain.MenuVersionClick(Sender: TObject);
begin
  FormAbout.Visible := True;
end; {procedure TFormMain.MenuVersionClick}

//------------------------------------------------------------------------------
procedure TFormMain.PlotBsols();
const MX = 10E20;
var
  b      : ^Tbootstrapsolution ;
  rh, ps : single;
  ib, i  : integer;
  mxb    : integer;
  x, y   : single;
  xs     : Tvector2; // mean
  xx     : Tvector2; // a bootstrap solution
  angl   : single;
  phs    : single;
  bxmm   : Tvector2; // min and max of x1
  bymm   : Tvector2; // min and max of x2
  brh    : Tvector2; // min and max of rho_s
  bps    : Tvector2; // min and max of psi_s
  tmp    : single;
  R95,ph95 : string;
begin
  xs[1]  := Dats.zt * cos(Dats.ps);
  xs[2]  := Dats.zt * sin(Dats.ps);
  xs := normalize_vector2(xs, tmp);
  if tmp=0 then begin
    Memo1.Lines.Add('Null strain.');
    Exit; // Mean vector is a zero vector. ====>
  end;

  bxmm[1] := MX; bxmm[2] := -MX;
  bymm[1] := MX; bymm[2] := -MX;
  brh[1]  := MX; brh[2]  := -MX;
  bps[1]  := MX; bps[2]  := -MX;
  for ib := 1 to NumBootstrap do begin
    New(b);
    b^.x :=  Bs[ib].x;
    x2rhopsi( b^.x, b^.rh, b^.ps );
    b^.ds := hyperbolic_dist(Bs[ib].x,Dats.x);
    Bsols.Add(b);
  end; {for}
  Bsols.Sort(CompareDistances);
  SeriesB.Clear();
  SeriesB.Visible := True;
  mxb := round(ConfidenceLevel * NumBootstrap);
  for ib := 0 to mxb - 1 do begin
    ps := Tbootstrapsolution(Bsols.Items[ib]^).ps;
    rh := Tbootstrapsolution(Bsols.Items[ib]^).rh;
    SeriesB.AddPolar( RadToDeg( ps ),  rho2zeta( rh ) );
    {estimate error bounds}
    x := rho2zeta(rh) * cos(ps);
    y := rho2zeta(rh) * sin(ps);
    xx[1] := x; xx[2] := y;
    xx := normalize_vector2(xx,tmp);
    if tmp = 0 then
      angl := 180
    else
      angl := radtodeg(arcsin( vecprod(xs,xx) )) / 2.0
    ; {end if}
    bxmm[1] := min(bxmm[1],x);   bxmm[2] := max(bxmm[2],x);
    bymm[1] := min(bymm[1],y);   bymm[2] := max(bymm[2],y);
    brh[1]  := min(brh[1],rh);   brh[2]  := max(brh[2],rh);
    bps[1]  := min(bps[1],angl); bps[2]  := max(bps[2],angl);
    if angl = 180 then bps[1] := -180;
  end; {end for}
  for ib := 0 to Bsols.Count - 1 do Dispose(Bsols.Items[ib]);
  Bsols.Clear();

  {estimate error bounds}
  if (bxmm[1]*bxmm[2] < 0) and (bymm[1]*bymm[2] < 0) then begin
    Memo1.Lines.Add('Confidence region surrounds the origin.');
    Exit; //                                                =====>
  end else begin
    Memo1.Lines.Add('');
    Memo1.Lines.Add('** 95% Confidence Limits **');
    R95 := '  R_s = ' + FloatToStrF(Dats.R, ffFixed,5,3) +
                    ' +' + FloatToStrF( Dats.R-exp(brh[1]),ffFixed,5,3) +
                    '/-' + FloatToStrF( exp(brh[2])-Dats.R,ffFixed,5,3);
    Memo1.Lines.Add(R95);
    Memo1.Lines.Add('      min = ' + FloatToStrF(exp(brh[1]),ffFixed,5,3) );
    Memo1.Lines.Add('      max = ' + FloatToStrF(exp(brh[2]),ffFixed,5,3) );
    phs := Dats.ph + 360 mod 360;

    ph95 := '  phi_s = ' + FloatToStrF(Dats.ph,ffFixed,5,3) + ' +' +
                           FloatToStrF(bps[2], ffFixed,5,3) +  '/' +
                           FloatToStrF(bps[1], ffFixed,5,3);
    Memo1.Lines.Add(ph95);
    Memo1.Lines.Add('      min = ' + FloatToStrF(Dats.ph+bps[1],ffFixed,5,3) );
    Memo1.Lines.Add('      max = ' + FloatToStrF(Dats.ph+bps[2],ffFixed,5,3) );
    StatusBar1.Panels[1].Text := R95 + '  ' + ph95;
  end; {if}

end; {procedure TFormMain.PlotBsols}

//------------------------------------------------------------------------------
procedure TFormMain.PlotDatF();
var i : integer;
begin
  SeriesF.Clear();
  for i := 1 to Ndat do SeriesF.AddPolar(RadToDeg(DatF[i].ps),DatF[i].zt);
  Chart1.LeftAxis.Maximum := R2zeta(FRmax())+0.5;
end; {procedure TFormMain.PlotDatF}

//------------------------------------------------------------------------------
procedure TFormMain.PlotDatI();
var i : integer;
begin
  SeriesI.Clear();
  for i := 1 to Ndat do SeriesI.AddPolar(RadToDeg(DatI[i].ps),DatI[i].zt);
end; {procedure TFormMain.PlotDatI}

//------------------------------------------------------------------------------
function TFormMain.ReadData(): boolean;
var
  f : TextFile;
  i : integer;
  sa, sb, st: single;
begin
  Result := False;
  if OpenDialog1.Execute() then begin
    if not FileExists(OpenDialog1.FileName) then begin
      ShowMessage('File not exits.');
      Exit;                              //===>
    end; {if}
    try
      AssignFile(f,OpenDialog1.FileName);
      Reset(f);
      i := 0;
      try
        while not Eof(f) do begin
          inc(i);
          Readln(f, sa, sb, st);
          if sa < sb then begin
            ShowMessage('Major axis is shorter than minor one.  #'+IntToStr(i));
            Exit;                        //===>
          end; {if}
          if sa < 0 then begin
            ShowMessage('Neative radii.  #'+IntToStr(i));
            Exit;                        //===>
          end; {if}
        end; {while}
      except
        ShowMessage('Read error --- #'+IntToStr(i));
        CloseFile(f);
        Exit;                            //===>
      end; {try}

      Ndat := i;
      StatusBar1.Panels[0].Text := 'N : '+IntToStr(Ndat);
      Finalize( DatF);         Finalize( DatI);
      SetLength(DatF, Ndat+1); SetLength(DatI, Ndat+1);
      StringGrid1.CleanupInstance();
      StringGrid1.RowCount := Ndat+1;
      Reset(f);
      for i := 1 to Ndat do begin
        Readln(f, DatF[i].a, DatF[i].b, DatF[i].ph);
        DatF[i].R := DatF[i].a / DatF[i].b;
        // DatF[i].ps := DegToRad( 2 * DatF[i].ph);
        StringGrid1.Cells[0,i] := IntToStr(i);
        StringGrid1.Cells[1,i] := FloatToStrF(DatF[i].a, ffFixed,5,2);
        StringGrid1.Cells[2,i] := FloatToStrF(DatF[i].b, ffFixed,5,2);
        StringGrid1.Cells[3,i] := FloatToStrF(DatF[i].ph,ffFixed,5,2);
        StringGrid1.Cells[4,i] := FloatToStrF(DatF[i].R, ffFixed,5,2);
      end;{for}
    finally
      CloseFile(f);
    end; {try}
  end; {if}

  Memo1.Lines.Clear();
  Memo1.Lines.Add('*** Data ***');
  Memo1.Lines.Add('  file = '+OpenDialog1.FileName);
  Memo1.Lines.Add('  N    = '+IntToStr(Ndat));
  SeriesFh.Visible := False;
  SeriesIh.Visible := False;
  FormMain.Caption := 'hStrain  [ ' + ExtractFileName(OpenDialog1.FileName) + ' ]';

  Result := True;
end; {function TFormMain.ReadData}

//------------------------------------------------------------------------------
procedure TFormMain.Rfphi2hypcoords(const k: AnsiChar);
var
  i : integer;
begin
  case k of
    'F' : for i := 1 to Ndat do begin
           DatF[i].rh := ln(DatF[i].R);
           DatF[i].ps := DegToRad(2* DatF[i].ph);
           DatF[i].zt := R2zeta(DatF[i].R);
          end; {F & for}
    'I' : for i := 1 to Ndat do begin
           DatI[i].rh := ln(DatI[i].R);
           DatI[i].ps := 2* DatI[i].ph;
           DatI[i].zt := R2zeta(DatI[i].R);
          end; {I & for}
  end; {case}
end; {procedure TFormMain.Rfphi2hypcoords}

//------------------------------------------------------------------------------
procedure TFormMain.SeriesAllClear();
begin
  SeriesB.Clear();
  SeriesF.Clear();
  SeriesI.Clear();
  SeriesO.Clear();
end; {procedure TFormMain.SeriesAllClear}

//------------------------------------------------------------------------------
procedure TFormMain.StringGrid1MouseDown(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
var
  arow, acol : integer;
begin
  if Ndat <= 0 then Exit; //===>

  StringGrid1.MouseToCell(x,y,acol,arow);

  SeriesFh.Clear();
  SeriesFh.Visible := True;
  SeriesFh.Transparency := 50;
  SeriesFh.AddPolar(RadToDeg(DatF[arow].ps),DatF[arow].zt);
  StatusBar1.Panels[1].Text := FloatToStrF(DatF[arow].R,ffFixed,5,2);

  SeriesIh.Clear();
  SeriesIh.Visible := True;
  Seriesih.Transparency := 50;
  SeriesIh.AddPolar(RadToDeg(DatI[arow].ps),DatI[arow].zt);
  StatusBar1.Panels[1].Text := FloatToStrF(DatI[arow].R,ffFixed,5,2);
end; {procedure TFormMain.StringGrid1MouseDown}

//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
end.
